/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2016 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.spiderchart.widget;

import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.fortiss.tooling.spiderchart.model.DataSeries;
import org.fortiss.tooling.spiderchart.model.SpiderChart;
import org.fortiss.tooling.spiderchart.style.ChartStyle;
import org.fortiss.tooling.spiderchart.style.FontStyle;
import org.fortiss.tooling.spiderchart.style.LegendStyle;
import org.fortiss.tooling.spiderchart.style.LineStyle;

/**
 * Class representing the spiderchart's legend.
 *
 * @author mondal
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating GREEN Hash: 13207B1B51F8F60D2122999D2EDB0603
 */
public final class SpiderChartLegendWidget extends SpiderChartWidgetBase {

	/** The size of the color icon of data series. */
	private static final int ICON_SIZE = 20;
	/** The space between icon and text. */
	private static final int ICON_SPACE = 3;

	/** Constructor. */
	public SpiderChartLegendWidget(SpiderChart chart, ChartStyle style) {
		super(chart, style);
	}

	/** Draws the provided graphics */
	public void draw(GC gc) {
		if(!style.isShowLegend()) {
			return;
		}

		if(style.getLegendStyle().isVerticalLayout()) {
			drawVertical(gc);
		} else {
			drawHorizontal(gc);
		}
	}

	/** draws the graphics horizontally */
	public void drawHorizontal(GC gc) {
		String labelText = chart.getLegendLabel();
		int margin = style.getLegendStyle().getMargin();

		int xDelta = 0;
		if(labelText != null && labelText.trim().length() > 0) {
			xDelta = drawHorizontalLabel(gc);
		}

		Color oldColor = gc.getForeground();
		int y = getY() + margin;
		for(DataSeries data : chart.getDataSeries()) {
			String text = data.getName();
			Point extent = gc.textExtent(text);
			LineStyle l = style.getDataSeriesStyle(data).getLineStyle();
			Color color = l.createSWTColor();
			gc.setForeground(color);

			int xl = getX() + xDelta;
			l.draw(gc, xl, y + extent.y / 2, xl + ICON_SIZE, y + extent.y / 2);
			xDelta = xDelta + ICON_SIZE + ICON_SPACE;

			int xs = getX() + xDelta;
			gc.drawString(text, xs, y, true);
			xDelta = xDelta + extent.x + margin;
			color.dispose();
		}
		gc.setForeground(oldColor);
	}

	/** Draws the horizontal legend label and returns the offset. */
	private int drawHorizontalLabel(GC gc) {
		String labelText = chart.getLegendLabel();
		LegendStyle legendStyle = style.getLegendStyle();
		FontStyle labelStyle = legendStyle.getLabelStyle();
		int margin = legendStyle.getMargin();

		Font oldFont = gc.getFont();
		Color oldColor = gc.getForeground();

		Font font = labelStyle.createSWTFont();
		gc.setFont(font);
		Color color = labelStyle.createSWTColor();
		gc.setForeground(color);
		gc.setClipping(getX(), getY(), getWidth(), getHeight());

		Point textExtent = gc.textExtent(labelText);
		gc.drawString(labelText, getX() + margin, getY() + margin, true);

		gc.setClipping((Rectangle)null);
		gc.setForeground(oldColor);
		color.dispose();
		gc.setFont(oldFont);
		font.dispose();

		return textExtent.x + 2 * margin;
	}

	/** draws the graphics vertically */
	public void drawVertical(GC gc) {
		String labelText = chart.getLegendLabel();
		LegendStyle legendStyle = style.getLegendStyle();
		int margin = legendStyle.getMargin();

		int yDelta = 0;
		if(labelText != null && labelText.trim().length() > 0) {
			yDelta = drawVerticalLabel(gc);
		}

		int textWidth = 0;
		int textHeight = 0;

		for(DataSeries data : chart.getDataSeries()) {
			Point extent = gc.textExtent(data.getName());
			textWidth = Math.max(textWidth, extent.x);
			textHeight = Math.max(textHeight, extent.y);
		}

		Color oldColor = gc.getForeground();
		int cnt = 0;
		for(DataSeries data : chart.getDataSeries()) {
			int dY = yDelta + cnt * (textHeight + margin);
			int dX = ICON_SIZE + 2 * margin;
			LineStyle l = style.getDataSeriesStyle(data).getLineStyle();

			Color color = l.createSWTColor();
			gc.setForeground(color);
			int lY = dY + textHeight / 2;
			l.draw(gc, getX() + margin, getY() + lY, getX() + margin + ICON_SIZE, getY() + lY);
			gc.drawText(data.getName(), getX() + dX, getY() + dY, true);
			color.dispose();

			cnt++;

		}
		gc.setForeground(oldColor);
	}

	/** Draws the header label for the vertical legend layout and returns the offset. */
	private int drawVerticalLabel(GC gc) {
		String labelText = chart.getLegendLabel();
		LegendStyle legendStyle = style.getLegendStyle();
		FontStyle labelStyle = legendStyle.getLabelStyle();
		int margin = legendStyle.getMargin();

		Font oldFont = gc.getFont();
		Color oldColor = gc.getForeground();

		Font font = labelStyle.createSWTFont();
		gc.setFont(font);
		Color color = labelStyle.createSWTColor();
		gc.setForeground(color);
		gc.setClipping(getX(), getY(), getWidth(), getHeight());

		Point textExtent = gc.textExtent(labelText);
		int x = getX() + margin + (getWidth() - margin - textExtent.x) / 2;
		gc.drawString(labelText, x, getY() + margin, true);

		gc.setClipping((Rectangle)null);
		gc.setForeground(oldColor);
		color.dispose();
		gc.setFont(oldFont);
		font.dispose();

		return textExtent.y + margin;
	}
}
