/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2016 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.spiderchart.widget;

import static org.fortiss.tooling.spiderchart.util.AxisUtils.getSegmentPoint;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.FontData;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Path;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.widgets.Display;
import org.fortiss.tooling.spiderchart.model.Axis;
import org.fortiss.tooling.spiderchart.model.DataSeries;
import org.fortiss.tooling.spiderchart.model.SpiderChart;
import org.fortiss.tooling.spiderchart.style.AxisStyle;
import org.fortiss.tooling.spiderchart.style.ChartStyle;
import org.fortiss.tooling.spiderchart.style.DataSeriesStyle;
import org.fortiss.tooling.spiderchart.style.FillStyle;
import org.fortiss.tooling.spiderchart.style.FontStyle;

/**
 * Class for drawing the spider chart.
 *
 * @author mondal
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating GREEN Hash: 13207B1B51F8F60D2122999D2EDB0603
 */
public final class SpiderChartWidget extends SpiderChartWidgetBase {
	/** Constructor. */
	public SpiderChartWidget(SpiderChart chart, ChartStyle style) {
		super(chart, style);
	}

	/** Draws the spider chart. */
	public void draw(GC gc) {
		compute();
		if(!style.isUseIndividualAxisSegments()) {
			drawBackground(gc);
			drawWebLines(gc);
		}
		drawAxes(gc);
		for(DataSeries data : chart.getDataSeries()) {
			drawDataSeries(gc, data);
		}
	}

	/** The center point of the chart. */
	private Point center = new Point(0, 0);
	/** The extent in pixels from the center point. */
	private int extent = 0;
	/** The end points of the axes. */
	private Map<Axis, Point> axesEnds = new HashMap<Axis, Point>();
	/** The angle in degree between two axes. */
	private double segmentDegree = 0.0;

	/** Computes the end point for each axis. */
	// FIXME: do not call this during draw, but when widget size changes.
	private void compute() {
		List<Axis> axes = chart.getAxes();
		int size = axes.size();
		segmentDegree = 360.0 / size;
		double currentAngleDegree = style.getStartAngleDegree();

		int halfWidth = getWidth() / 2;
		int halfHeight = getHeight() / 2;
		center.x = getX() + halfWidth;
		center.y = getY() + halfHeight;

		double relativeMargin = 1.0 - style.getRelativeMargin();
		int wExtent = (int)(relativeMargin * halfWidth);
		int hExtent = (int)(relativeMargin * halfHeight);
		extent = Math.min(wExtent, hExtent);

		for(int i = 0; i < size; i++) {
			Axis axis = axes.get(i);
			// compute end point
			double startAngleRad = Math.PI * currentAngleDegree / 180.0;
			int extentX = (int)(Math.cos(startAngleRad) * extent);
			int extentY = (int)(Math.sin(startAngleRad) * extent);
			Point outer = new Point(center.x + extentX, center.y + extentY);
			axesEnds.put(axis, outer);
			currentAngleDegree = (currentAngleDegree + segmentDegree) % 360.0;
		}
	}

	/** Draws the background if style requires all axes to have equal number of segments. */
	private void drawBackground(GC gc) {
		Path path = new Path(Display.getCurrent());
		boolean first = true;
		for(Axis axis : chart.getAxes()) {
			Point p = axesEnds.get(axis);
			if(first) {
				path.moveTo(p.x, p.y);
				first = false;
			} else {
				path.lineTo(p.x, p.y);
			}
		}
		path.close();

		Color oldBGColor = gc.getBackground();
		Color fillColor = style.getBackgroundFillStyle().createSWTColor();
		gc.setBackground(fillColor);
		gc.fillPath(path);
		gc.setBackground(oldBGColor);
		fillColor.dispose();

		path.dispose();
	}

	/** Draws the web lines of the background if all axes have equal number of segments. */
	private void drawWebLines(GC gc) {
		Color oldColor = gc.getForeground();
		Color color = style.getBackgroundLineStyle().createSWTColor();
		gc.setForeground(color);

		for(int i = 1; i <= style.getAxisSegments(); i++) {
			Path p = computePath(i);
			gc.drawPath(p);
			p.dispose();
		}

		gc.setForeground(oldColor);
		color.dispose();
	}

	/** Computes the i-th web line path. */
	private Path computePath(int i) {
		double ratio = (double)i / (double)style.getAxisSegments();
		boolean first = true;
		Path p = new Path(Display.getCurrent());
		for(Axis axis : chart.getAxes()) {
			Point outer = axesEnds.get(axis);
			int xExtent = outer.x - center.x;
			int yExtent = outer.y - center.y;
			int x = center.x + (int)(ratio * xExtent);
			int y = center.y + (int)(ratio * yExtent);
			if(first) {
				p.moveTo(x, y);
				first = false;
			} else {
				p.lineTo(x, y);
			}
		}
		p.close();
		return p;
	}

	/** Draws the data series. */
	private void drawDataSeries(GC gc, DataSeries data) {
		DataSeriesStyle dStyle = style.getDataSeriesStyle(data);
		Path path = new Path(Display.getCurrent());
		boolean first = true;
		for(Axis axis : chart.getAxes()) {
			Point p = computeDataPoint(axis, data);
			if(first) {
				path.moveTo(p.x, p.y);
				first = false;
			} else {
				path.lineTo(p.x, p.y);
			}
			if(dStyle.isShowIndicators()) {
				drawIndicator(gc, dStyle, p);
			}
			if(dStyle.isShowIndicatorLabels()) {
				drawIndicatorLabel(gc, dStyle, p, data.getAxisValue(axis));
			}
		}
		path.close();

		// draw the filled area with alpha value.
		FillStyle fStyle = dStyle.getFillStyle();
		Color oldBGColor = gc.getBackground();
		int oldAlpha = gc.getAlpha();
		gc.setAlpha(fStyle.getAlpha());
		Color fillColor = fStyle.createSWTColor();
		gc.setBackground(fillColor);
		gc.fillPath(path);
		gc.setBackground(oldBGColor);
		gc.setAlpha(oldAlpha);
		fillColor.dispose();

		// draw the outline without alpha value
		Color oldFGColor = gc.getForeground();
		Color outlineColor = dStyle.getLineStyle().createSWTColor();
		gc.drawPath(path);
		gc.setForeground(oldFGColor);
		outlineColor.dispose();

		path.dispose();
	}

	/** Draws the indicator labels next to the chart point. */
	private void drawIndicatorLabel(GC gc, DataSeriesStyle dStyle, Point p, Double value) {
		Font oldFont = gc.getFont();
		Font font = dStyle.getIndicatorLabelStyle().createSWTFont();
		gc.setFont(font);

		Color oldColor = gc.getForeground();
		Color color = dStyle.getIndicatorLabelStyle().createSWTColor();
		gc.setForeground(color);

		String label = dStyle.getDecimalFormat().format(value);
		Point textExtent = gc.textExtent(label);
		// data series indicator labels are always to the right of the point
		int x = p.x + 4;
		int y = p.y;
		if(y > center.y) {
			y = y + 4;
		} else if(y < center.y) {
			y = y - 4 - textExtent.y;
		} else {
			y = y - textExtent.y / 2;
		}
		gc.drawString(label, x, y, true);

		gc.setForeground(oldColor);
		color.dispose();
		gc.setFont(oldFont);
		font.dispose();
	}

	/** Draws the filled indicator rectangle. */
	private void drawIndicator(GC gc, DataSeriesStyle dStyle, Point p) {
		int size = dStyle.getIndicatorSize();
		Color oldColor = gc.getBackground();
		Color color = dStyle.getLineStyle().createSWTColor();
		gc.setBackground(color);
		gc.fillRectangle(p.x - size / 2, p.y - size / 2, size, size);
		gc.setBackground(oldColor);
		color.dispose();
	}

	/** Computes the point on the given axis for the given data. */
	private Point computeDataPoint(Axis axis, DataSeries data) {
		Double ratio = axis.getAxisRatio(data.getAxisValue(axis));
		Point outer = axesEnds.get(axis);
		int deltaX = (int)(ratio * (outer.x - center.x));
		int deltaY = (int)(ratio * (outer.y - center.y));
		return new Point(center.x + deltaX, center.y + deltaY);
	}

	/** Draws the axes of the spider chart. */
	private Map<Axis, Point> drawAxes(GC gc) {
		List<Axis> axes = chart.getAxes();
		int size = axes.size();
		double segmentDegree = 360.0 / size;
		double angleDegree = 90.0;

		int halfWidth = getWidth() / 2;
		int halfHeight = getHeight() / 2;
		int centerX = getX() + halfWidth;
		int centerY = getY() + halfHeight;

		for(int i = 0; i < size; i++) {
			Axis axis = axes.get(i);
			Point outer = axesEnds.get(axis);
			// draw line
			AxisStyle aStyle = style.getAxisStyle(axis);
			aStyle.getLineStyle().draw(gc, centerX, centerY, outer.x, outer.y);
			// draw label
			drawAxisLabel(gc, axis, aStyle, centerX, centerY, outer.x, outer.y);
			// draw segment indicators
			int segments =
					style.isUseIndividualAxisSegments() ? aStyle.getSegments() : style
							.getAxisSegments();
			for(int s = 1; s <= segments; s++) {
				double ratio = (double)s / (double)segments;
				String lbl = aStyle.getDecimalFormat().format(axis.getAxisValue(ratio));
				Point segmentPoint =
						getSegmentPoint(segments, s, outer.x - center.x, outer.y - center.y);
				drawSegmentIndicator(gc, lbl, aStyle, centerX, centerY, segmentPoint);
			}

			angleDegree = (angleDegree + segmentDegree) % 360.0;
		}
		return axesEnds;
	}

	/** Draws the indicator on the axis. */
	private void drawSegmentIndicator(GC gc, String lbl, AxisStyle axisStyle, int centerX,
			int centerY, Point segment) {
		FontStyle segmentStyle = axisStyle.getSegmentStyle();
		Font oldFont = gc.getFont();
		Font font = segmentStyle.createSWTFont();
		gc.setFont(font);

		Color oldColor = gc.getForeground();
		Color color = segmentStyle.createSWTColor();
		gc.setForeground(color);

		// segment indicator labels are always to the left of the axis
		int posX = centerX + segment.x + -gc.textExtent(lbl).x - 4;
		int posY = centerY + segment.y + 4;
		gc.drawText(lbl, posX, posY, true);

		gc.setForeground(oldColor);
		color.dispose();
		gc.setFont(oldFont);
		font.dispose();
	}

	/** Draws the axis label at the correct position. */
	private void drawAxisLabel(GC gc, Axis axis, AxisStyle style, int centerX, int centerY,
			int outerX, int outerY) {
		String axisLabel = axis.getName();
		FontData fontData = style.getLabelStyle().getFontData();

		Font oldFont = gc.getFont();
		Font font = new Font(Display.getCurrent(), fontData);
		gc.setFont(font);

		Color oldColor = gc.getForeground();
		Color color = style.getLabelStyle().createSWTColor();
		gc.setForeground(color);

		int textExtent = gc.textExtent(axisLabel).x;
		if(outerX > centerX) {
			outerX += 5;
		} else if(outerX < centerX) {
			outerX = outerX - 5 - textExtent;
		} else {
			outerX = outerX - textExtent / 2;
		}
		if(outerY > centerY) {
			outerY = outerY + 5;
		} else if(outerY < centerY) {
			outerY = outerY - 5 - fontData.getHeight();
		} else {
			outerY = outerY - fontData.getHeight() / 2;
		}
		gc.drawString(axisLabel, outerX, outerY, true);

		gc.setFont(oldFont);
		font.dispose();
		gc.setForeground(oldColor);
		color.dispose();
	}
}
